# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import logging
from mock import patch, call, MagicMock

from ...utils import swallow_logs
from ...tests.utils import assert_in
from ..base import ResourceManager
from ..ssh import SSH


def test_awsec2_class():

    with patch('boto3.resource') as client, \
            patch.object(SSH, 'get_session', return_value='started_session'), \
            swallow_logs(new_level=logging.DEBUG) as log:

        # Test connecting when a resource doesn't exist.
        client.return_value = MagicMock(
            instances=MagicMock(filter=lambda Filters: [])
        )
        config = {
            'name': 'non-existent-instance',
            'type': 'aws-ec2',
            'access_key_id': 'my-aws-access-key-id',
            'secret_access_key': 'my-aws-secret-access-key-id'
        }
        resource = ResourceManager.factory(config)
        resource.connect()
        assert resource.id is None
        assert resource.status is None

        # Test catching exception when multiple resources are found at connection.
        client.return_value = MagicMock(
            instances=MagicMock(filter=lambda Filters: [
                MagicMock(
                    instance_id='i-22221ddf096c22bb0',
                    state={'Name': 'running'}
                ),
                MagicMock(
                    instance_id='i-3333f40de2b9b8967',
                    state={'Name': 'running'}
                )
            ])
        )
        config = {
            'name': 'duplicate-instance-name',
            'type': 'aws-ec2',
            'access_key_id': 'my-aws-access-key-id',
            'secret_access_key': 'my-aws-secret-access-key-id'
        }
        resource = ResourceManager.factory(config)
        try:
            resource.connect()
        except Exception as e:
            assert e.args[0] == "Multiple container matches found"

        # Test connecting to an existing resource.
        client.return_value = MagicMock(
            Instance=lambda id: MagicMock(
                instance_id='i-00002777d52482d9c',
                state={'Name': 'running'}
            )
        )
        config = {
            'name': 'my-instance-name',
            'id': 'i-00002777d52482d9c',
            'type': 'aws-ec2',
            'access_key_id': 'my-aws-access-key-id',
            'secret_access_key': 'my-aws-secret-access-key-id',
            'key_name': 'my-ssh-key',
            'key_filename': '/home/me/.ssh/id_rsa'
        }
        resource = ResourceManager.factory(config)
        resource.connect()
        assert resource.image == 'ami-c8580bdf'
        assert resource.id == 'i-00002777d52482d9c'
        assert resource.instance_type == 't2.micro'
        assert resource.key_filename == '/home/me/.ssh/id_rsa'
        assert resource.key_name == 'my-ssh-key'
        assert resource.name == 'my-instance-name'
        assert resource.region_name == 'us-east-1'
        assert resource.access_key_id == 'my-aws-access-key-id'
        assert resource.secret_access_key == 'my-aws-secret-access-key-id'
        assert resource.security_group == 'default'
        assert resource.status == 'running'
        assert resource.type == 'aws-ec2'

        # Test creating an existing resource and catch the exception.
        try:
            resource.create()
        except Exception as e:
            assert e.args[0] == "Instance 'i-00002777d52482d9c' already exists in AWS subscription"

        # Test creating resource.
        client.return_value = MagicMock(
            instances=MagicMock(filter=lambda Filters: []),
            Instance=lambda id: MagicMock(
                instance_id='i-11112777d52482d9c',
                state={'Name': 'running'}
            )
        )
        config = {
            'name': 'my-instance-name',
            'type': 'aws-ec2',
            'access_key_id': 'my-aws-access-key-id',
            'secret_access_key': 'my-aws-secret-access-key-id',
            'key_name': 'my-ssh-key',
            'key_filename': '/home/me/.ssh/id_rsa'
        }
        resource = ResourceManager.factory(config)
        resource.connect()
        results = resource.create()
        assert results['id'] == 'i-11112777d52482d9c'
        assert results['status'] == 'running'
        assert_in('EC2 instance i-11112777d52482d9c initialized!', log.lines)
        assert_in('EC2 instance i-11112777d52482d9c to start running!', log.lines)
        assert_in('Waiting for EC2 instance i-11112777d52482d9c to complete initialization...', log.lines)
        assert_in('EC2 instance i-11112777d52482d9c initialized!', log.lines)

        # Test stopping resource.
        resource.stop()
        calls = [
            call.stop()
        ]
        resource._ec2_instance.assert_has_calls(calls, any_order=True)

        # Test starting resource.
        resource.start()
        calls = [
            call.start()
        ]
        resource._ec2_instance.assert_has_calls(calls, any_order=True)

        # Test deleting resource.
        resource.delete()
        calls = [
            call.terminate()
        ]
        resource._ec2_instance.assert_has_calls(calls, any_order=True)

        session = resource.get_session()
        assert session == 'started_session'



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import logging
import mock
from six.moves import StringIO

from niceman.api import backend_parameters
from niceman.utils import swallow_logs


def test_backend_parameters_unknown_resource():
    with swallow_logs(new_level=logging.WARNING) as log:
        backend_parameters(["i'm-unknown"])
        assert "Failed to import" in log.out


def test_backend_parameters_with_arg():
    with mock.patch('sys.stdout', new_callable=StringIO) as out:
        backend_parameters(["ssh"])
    assert "host:" in out.getvalue()
    assert "aws_ec2" not in out.getvalue()


def test_backend_parameters_all():
    with mock.patch('sys.stdout', new_callable=StringIO) as out:
        backend_parameters()
    assert "host:" in out.getvalue()
    assert "aws_ec2" in out.getvalue()



# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Test sequences of interface commands.
"""

import logging
import os
import os.path as op
import pytest

from mock import patch
import yaml

from niceman.cmdline.main import main
from niceman.cmd import Runner
from niceman.resource.base import ResourceManager
from niceman.support.exceptions import CommandError
from niceman.tests.utils import swallow_logs


def test_create_start_stop(tmpdir):
    tmpdir = str(tmpdir)
    inventory_file = op.join(tmpdir, "inventory.yml")
    rm = ResourceManager(inventory_file)

    # Simple smoke test.  We can't easily test the effects of start/stop with
    # shell because those the start and stop methods are noops.
    with patch("niceman.interface.create.get_manager",
               return_value=rm):
        main(["create", "-t", "shell", "testshell"])

    with open(inventory_file) as ifh:
        inventory = yaml.safe_load(ifh)
    assert inventory["testshell"]["status"] == "N/A"

    with patch("niceman.interface.start.get_manager",
               return_value=rm):
        main(["start", "testshell"])

    with patch("niceman.interface.stop.get_manager",
               return_value=rm):
        main(["stop", "testshell"])

    with patch("niceman.interface.delete.get_manager",
               return_value=rm):
        main(["delete", "--skip-confirmation", "testshell"])

    with open(inventory_file) as ifh:
        inventory = yaml.safe_load(ifh)
    assert "testshell" not in inventory


@pytest.mark.integration
def test_create_and_start(tmpdir):
    runner = Runner()
    tmpdir = str(tmpdir)
    cfg_file = op.join(tmpdir, "custom.cfg")
    inventory_file = op.join(tmpdir, "inventory.yml")
    with open(cfg_file, "w") as cfg_fh:
        cfg_fh.write("[general]\ninventory_file = {}\n".format(inventory_file))

    def run_niceman(args):
        runner(["niceman", "--config", cfg_file] + args,
               expect_stderr=True)

    run_niceman(["create", "--resource-type=shell", "myshell"])

    with open(inventory_file) as ifh:
        dumped = ifh.read()
    assert "myshell" in dumped
    assert "id" in dumped

    # Running with a different config fails ...
    empty_cfg_file = op.join(tmpdir, "empty.cfg")
    with open(empty_cfg_file, "w"):
        pass

    with swallow_logs(new_level=logging.ERROR) as cml:
        with pytest.raises(CommandError):
            runner(["niceman", "--config", empty_cfg_file,
                    "start", "myshell"])
        if os.environ.get("NICEMAN_LOGTARGET", "stderr") == "stderr":
            assert "ResourceNotFoundError" in cml.out
    # ... but using the same config works.
    run_niceman(["start", "myshell"])



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""NICEMAN - Neuroimaging Computation Environments Manager

It is a part of the ReproNim (Reproducible Neuroimaging) suite of tools
to ease construction and execution of computation environments based on the
provenance data.
"""

from __future__ import absolute_import

from .log import lgr

# Other imports are interspersed with lgr.debug to ease troubleshooting startup
# delays etc.
lgr.log(5, "Instantiating config")
from .config import ConfigManager
cfg = ConfigManager()

# Not used (for now)
# lgr.log(5, "Instantiating ssh manager")
# from .support.sshconnector import SSHManager
# ssh_manager = SSHManager()
#
import atexit
import niceman
import os
# atexit.register(ssh_manager.close, allow_fail=False)
atexit.register(lgr.log, 5, "Exiting")

from .version import __version__


def test(package='niceman', **kwargs):
    """A helper to run niceman's tests.  Requires numpy and pytest

    See numpy.testing.Tester -- **kwargs are passed into the
    Tester().test call
    """
    import pytest
    try:
        # from numpy.testing import Tester
        # Tester(package=package).test(**kwargs)
        pytest.main(['-s', '--disable-pytest-warnings', os.path.dirname(__file__)])
        # we don't have any benchmarks atm
        # bench = Tester().bench
    except ImportError:
        raise RuntimeError('Problem with pytest for niceman.tests().  Nothing is done')
test.__test__ = False

# Following fixtures are necessary at the top level __init__ for fixtures which
# would cover all **/tests and not just niceman/tests/

# To store settings which setup_package changes and teardown_package should return
_test_states = {
    'loglevel': None,
    'NICEMAN_LOGLEVEL': None,
}

def setup_package():
    import os

    # To overcome pybuild overriding HOME but us possibly wanting our
    # own HOME where we pre-setup git for testing (name, email)
    if 'GIT_HOME' in os.environ:
        os.environ['HOME'] = os.environ['GIT_HOME']

    # To overcome pybuild by default defining http{,s}_proxy we would need
    # to define them to e.g. empty value so it wouldn't bother touching them.
    # But then haskell libraries do not digest empty value nicely, so we just
    # pop them out from the environment
    for ev in ('http_proxy', 'https_proxy'):
        if ev in os.environ and not (os.environ[ev]):
            lgr.debug("Removing %s from the environment since it is empty", ev)
            os.environ.pop(ev)

    NICEMAN_LOGLEVEL = os.environ.get('NICEMAN_LOGLEVEL', None)
    if NICEMAN_LOGLEVEL is None:
        # very very silent.  Tests introspecting logs should use
        # swallow_logs(new_level=...)
        _test_states['loglevel'] = lgr.getEffectiveLevel()
        lgr.setLevel(100)

        # And we should also set it within environ so underlying commands also stay silent
        _test_states['NICEMAN_LOGLEVEL'] = NICEMAN_LOGLEVEL
        os.environ['NICEMAN_LOGLEVEL'] = '100'
    else:
        # We are not overriding them, since explicitly were asked to have some log level
        _test_states['loglevel'] = None


def teardown_package():
    import os
    if os.environ.get('NICEMAN_TESTS_NOTEARDOWN'):
        return

    if _test_states['loglevel'] is not None:
        lgr.setLevel(_test_states['loglevel'])
        if _test_states['NICEMAN_LOGLEVEL'] is None:
            os.environ.pop('NICEMAN_LOGLEVEL')
        else:
            os.environ['NICEMAN_LOGLEVEL'] = _test_states['NICEMAN_LOGLEVEL']

    from niceman.tests import _TEMP_PATHS_GENERATED
    from niceman.tests.utils import rmtemp
    if len(_TEMP_PATHS_GENERATED):
        msg = "Removing %d dirs/files: %s" % (len(_TEMP_PATHS_GENERATED), ', '.join(_TEMP_PATHS_GENERATED))
    else:
        msg = "Nothing to remove"
    lgr.debug("Teardown tests. " + msg)
    for path in _TEMP_PATHS_GENERATED:
        rmtemp(path, ignore_errors=True)

lgr.log(5, "Done importing main __init__")



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import docker
import pytest

from ...distributions.docker import DockerDistribution
from ...distributions.docker import DockerImage
from ...distributions.docker import DockerTracer
from ...resource.session import get_local_session
from ...support.exceptions import CommandError
from ...tests.utils import skip_if_no_docker_engine, skip_if_no_network


@skip_if_no_network
@skip_if_no_docker_engine
def test_docker_trace_tag():
    client = docker.Client()
    client.pull('alpine:3.6')

    tracer = DockerTracer()

    files = ['alpine:3.6', 'non-existent-image']
    dist, remaining_files = next(tracer.identify_distributions(files))
    assert dist.name == 'docker'
    assert dist.images[0].architecture == 'amd64'
    assert dist.images[0].operating_system == 'linux'
    assert dist.images[0].repo_digests[0].startswith('alpine@sha256:')
    assert dist.images[0].repo_tags[0] == 'alpine:3.6'
    assert 'non-existent-image' in remaining_files


@skip_if_no_network
@skip_if_no_docker_engine
def test_docker_trace_id():
    client = docker.Client()
    repo_id = 'sha256:f625bd3ff910ad2c68a405ccc5e294d2714fc8cfe7b5d80a8331c72ad5cc7630'
    name = 'alpine@' + repo_id
    client.pull(name)

    tracer = DockerTracer()

    files = [name]
    dist, remaining_files = next(tracer.identify_distributions(files))
    assert dist.name == 'docker'
    assert dist.images[0].id == 'sha256:77144d8c6bdce9b97b6d5a900f1ab85da' + \
        '325fe8a0d1b0ba0bbff2609befa2dda'
    assert dist.images[0].architecture == 'amd64'
    assert dist.images[0].operating_system == 'linux'
    assert dist.images[0].id == 'sha256:77144d8c6bdce9b97b6d5a900f1ab85da' + \
        '325fe8a0d1b0ba0bbff2609befa2dda'
    assert dist.images[0].created == '2018-01-09T21:10:38.538173323Z'


@skip_if_no_network
@skip_if_no_docker_engine
def test_docker_trace_local_image():
    client = docker.Client()
    client.pull('alpine:3.6')
    tracer = DockerTracer()
    # Test tracing a local image not saved in a repository
    container = client.create_container(image='alpine:3.6',
        command='echo foo > test.txt')
    new_image = client.commit(container, repository='test-container',
        tag="001")
    client.remove_container(container)
    tracer = DockerTracer()
    files = [new_image['Id']]
    dist, _ = next(tracer.identify_distributions(files))
    assert dist.name == 'docker'
    assert dist.images[0].repo_tags[0] == 'test-container:001'

    # Clean up docker engine
    client.remove_image(new_image['Id'])


@skip_if_no_network
@skip_if_no_docker_engine
def test_docker_distribution():

    client = docker.Client()
    session = get_local_session()

    # Verify alpine:3.5 image is not stored locally
    try:
        client.remove_image('sha256:6c6084ed97e5851b5d216b20ed185230127' +
            '8584c3c6aff915272b231593f6f98')
    except docker.errors.NotFound:
        pass

    # Test tracing valid images
    dist = DockerDistribution('docker')
    dist.images = [
        DockerImage(
            'sha256:6c6084ed97e5851b5d216b20ed1852301278584c3c6aff915272' +
                'b231593f6f98',
            repo_digests=['alpine@sha256:9148d069e50eee519ec45e5683e56a1' +
                'c217b61a52ed90eb77bdce674cc212f1e'],
            repo_tags=['alpine:3.5']
        ),
        DockerImage(
            'sha256:77144d8c6bdce9b97b6d5a900f1ab85da325fe8a0d1b0ba0bbff2' +
                '609befa2dda',
            repo_digests=['alpine@sha256:f625bd3ff910ad2c68a405ccc5e294d2' +
                '714fc8cfe7b5d80a8331c72ad5cc7630'],
            repo_tags=['alpine:3.6']
        )
    ]
    dist.initiate(session)
    dist.install_packages(session)
    alpine_3_5 = client.inspect_image(dist.images[0].id)
    assert alpine_3_5['Id'] == 'sha256:6c6084ed97e5851b5d216b20ed18523012' + \
        '78584c3c6aff915272b231593f6f98'
    assert 'alpine@sha256:9148d069e50eee519ec45e5683e56a1c217b61a52ed90eb' + \
        '77bdce674cc212f1e' in alpine_3_5['RepoDigests']
    alpine_3_6 = client.inspect_image(dist.images[1].id)
    assert alpine_3_6['Id'] == 'sha256:77144d8c6bdce9b97b6d5a900f1ab85da3' + \
        '25fe8a0d1b0ba0bbff2609befa2dda'

    # FIXME: Tag checks are disabled to avoid errors from unstable IDs.  We
    # should switch to using images that under our control.  See gh-254.
    #
    # assert 'alpine:3.5' in alpine_3_5['RepoTags']
    # assert 'alpine:3.6' in alpine_3_6['RepoTags']

    # Clean up docker engine
    client.remove_image(dist.images[0].id)
    client.remove_image(dist.images[1].id)

    # Test installing a non-existent image
    dist.images = [
        DockerImage(
            'sha256:000000000000000000000000000000000000000000000000000000',
            repo_digests=['foo@00000000000000000000000000000000000000000'],
            repo_tags=['foo:latest']
        )
    ]
    with pytest.raises(CommandError):
        dist.install_packages(session)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Resource sub-class to provide management of a SSH connection."""

import attr
import invoke
import uuid
from fabric import Connection

import logging
lgr = logging.getLogger('niceman.resource.ssh')

from .base import Resource
from ..utils import attrib
from ..utils import command_as_string
from niceman.dochelpers import borrowdoc
from niceman.resource.session import Session
from ..support.exceptions import CommandError


@attr.s
class SSH(Resource):

    # Generic properties of any Resource
    name = attrib(default=attr.NOTHING)

    # Configurable options for each "instance"
    host = attrib(default=attr.NOTHING, doc="DNS or IP address of server")
    port = attrib(
        doc="Port to connect to on remote host")
    key_filename = attrib(
        doc="Path to SSH private key file matched with AWS key name parameter")
    user = attrib(
        doc="Username to use to log into remote environment")

    id = attrib()  # EC2 instance ID

    type = attrib(default='ssh')  # Resource type

    # Current instance properties, to be set by us, not augmented by user
    status = attrib()
    _connection = attrib()

    def connect(self, password=None):
        """Open a connection to the environment resource.

        Parameters
        ----------
        password : string
            We don't allow users to pass in a password via the command line
            but do allow tests to authenticate by passing a password as
            a parameter to this method.
        """
        self._connection = Connection(
            self.host,
            user=self.user,
            port=self.port,
            connect_kwargs={
                'key_filename': self.key_filename,
                'password': password
            }
        )

        if self.key_filename:
            auth = self.key_filename
        elif password is None:
            auth = "SSH config"
        else:
            auth = "password"

        lgr.debug("SSH connecting to %s@%s:%s, authenticating with %s",
                  self._connection.user, self._connection.host,
                  self._connection.port,  # Fabric defaults to 22.
                  auth)

        self._connection.open()

    def create(self):
        """
        Register the SSH connection to the niceman inventory registry.

        Returns
        -------
        dict : config and state parameters to capture in the inventory file
        """
        if not self.id:
            self.id = str(uuid.uuid4())
        self.status = 'N/A'
        return {
            'id': self.id,
            'status': self.status,
            'host': self.host,
            'user': self.user,
            'port': self.port,
            'key_filename': self.key_filename,
        }

    def delete(self):
        self._connection = None
        return

    def start(self):
        # Not a SSH feature
        raise NotImplementedError

    def stop(self):
        # Not a SSH feature
        raise NotImplementedError

    def get_session(self, pty=False, shared=None):
        """
        Log into remote environment and get the command line
        """
        if not self._connection:
            self.connect()

        return (PTYSSHSession if pty else SSHSession)(
            connection=self._connection
        )


# Alias SSH class so that it can be discovered by the ResourceManager.
@attr.s
class Ssh(SSH):
    pass


from niceman.resource.session import POSIXSession


@attr.s
class SSHSession(POSIXSession):
    connection = attrib(default=attr.NOTHING)

    @borrowdoc(Session)
    def _execute_command(self, command, env=None, cwd=None):
        # TODO -- command_env is not used etc...
        # command_env = self.get_updated_env(env)
        if env:
            raise NotImplementedError("passing env variables to execution")

        if cwd:
            raise NotImplementedError("implement cwd support")
        # if command_env:
            # TODO: might not work - not tested it
            # command = ['export %s=%s;' % k for k in command_env.items()] + command

        command = command_as_string(command)
        try:
            result = self.connection.run(command, hide=True)
        except invoke.exceptions.UnexpectedExit as e:
            result = e.result

        if result.return_code not in [0, None]:
            msg = "Failed to run %r. Exit code=%d. out=%s err=%s" \
                % (command, result.return_code, result.stdout, result.stderr)
            raise CommandError(str(command), msg, result.return_code,
                result.stdout, result.stderr)
        else:
            lgr.log(8, "Finished running %r with status %s", command,
                result.return_code)

        return (result.stdout, result.stderr)

    @borrowdoc(Session)
    def put(self, src_path, dest_path, uid=-1, gid=-1):
        dest_path = self._prepare_dest_path(src_path, dest_path, local=False)
        self.connection.put(src_path, dest_path)

        if uid > -1 or gid > -1:
            self.chown(dest_path, uid, gid)

    @borrowdoc(Session)
    def get(self, src_path, dest_path=None, uid=-1, gid=-1):
        dest_path = self._prepare_dest_path(src_path, dest_path)
        self.connection.get(src_path, dest_path)

        if uid > -1 or gid > -1:
            self.chown(dest_path, uid, gid, remote=False)


@attr.s
class PTYSSHSession(SSHSession):
    """Interactive SSH Session"""

    @borrowdoc(Session)
    def open(self):
        lgr.debug("Opening TTY connection via SSH.")
        self.interactive_shell()

    @borrowdoc(Session)
    def close(self):
        # XXX ?
        pass

    def interactive_shell(self):
        """Open an interactive TTY shell.
        """
        self.connection.run('/bin/bash', pty=True)
        print('Exited terminal session.')



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from os.path import join as opj, dirname

# Sample output from Reprozip.
REPROZIP_SPEC1_YML_FILENAME = opj(dirname(__file__), 'files', 'reprozip_spec1.yml')
REPROZIP_SPEC2_YML_FILENAME = opj(dirname(__file__), 'files', 'reprozip_xeyes.yml')

# examples/demo_spec1.yml
NICEMAN_SPEC1_YML_FILENAME = opj(dirname(__file__), 'files', 'nicemanspec1_v0.0.1.yml')
NICEMAN_SPEC1_YML = open(NICEMAN_SPEC1_YML_FILENAME).read()




# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the niceman package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Resource sub-class to provide management of the localhost environment."""

import attr
import shutil

from .base import Resource
from niceman.cmd import Runner
from niceman.dochelpers import borrowdoc
from niceman.resource.session import Session
from niceman.support.exceptions import CommandError
from niceman.utils import attrib

import logging
lgr = logging.getLogger('niceman.resource.shell')

import os

from .session import POSIXSession, get_updated_env


# For now just assuming that local shell is a POSIX shell
# Later we could specialize based on the OS, and that is why
# Resource/Shell is not subclassing Session but rather delegates to .session
class ShellSession(POSIXSession):
    """Local shell session"""

    def __init__(self):
        super(ShellSession, self).__init__()
        self._runner = None

    @borrowdoc(Session)
    def open(self):
        self._runner = Runner()

    @borrowdoc(Session)
    def close(self):
        self._runner = None

    @borrowdoc(Session)
    def _execute_command(self, command, env=None, cwd=None):
        # XXX should it be a generic behavior to auto-start?
        if self._runner is None:
            self.open()
        run_kw = {}
        if env:
            # if anything custom, then we need to get original full environment
            # and update it with custom settings which we either "accumulated"
            # via set_envvar, or it was passed into this call.
            run_kw['env'] = get_updated_env(os.environ, env)

        return self._runner.run(
            command,
            # For now we do not ERROR out whenever command fails or provides
            # stderr -- analysis will be done outside
            expect_fail=True,
            expect_stderr=True,
            cwd=cwd,
            **run_kw
        )  # , shell=True)

    @borrowdoc(Session)
    def isdir(self, path):
        return os.path.isdir(path)

    @borrowdoc(Session)
    def mkdir(self, path, parents=False):
        if not os.path.exists(path):
            if parents:
                os.makedirs(path)
            else:
                try:
                    os.mkdir(path)
                except OSError:
                    raise CommandError(
                        msg="Failed to make directory {}".format(path))

    @borrowdoc(Session)
    def get(self, src_path, dest_path=None, uid=-1, gid=-1):
        dest_path = self._prepare_dest_path(src_path, dest_path)
        shutil.copy(src_path, dest_path)
        if uid > -1 or gid > -1:
            self.chown(dest_path, uid, gid, recursive=True)

    @borrowdoc(Session)
    def put(self, src_path, dest_path, uid=-1, gid=-1):
        # put is the same as get for the shell resource
        self.get(src_path, dest_path, uid, gid)

@attr.s
class Shell(Resource):

    # Container properties
    name = attrib(default=attr.NOTHING)
    id = attrib()
    type = attrib(default='shell')

    status = attrib()

    def create(self):
        """
        Create a running environment.

        Parameters
        ----------
        name : string
            Name identifier of the environment to be created.
        """
        # Generic logic to reside in Resource???
        if self.id is None:
            self.id = Resource._generate_id()
        return {
            'id': self.id,
            'status': 'N/A'
        }

    def connect(self):
        """
        Connect to an existing environment.
        """
        return self

    def delete(self):
        """
        Remove this environment from the backend.
        """
        return

    def start(self):
        """
        Start this environment in the backend.
        """
        # Not a shell feature
        raise NotImplementedError

    def stop(self):
        """
        Stop this environment in the backend.
        """
        # Not a shell feature
        raise NotImplementedError

    def get_session(self, pty=False, shared=None):
        """
        Log into a container and get the command line
        """
        if pty:
            raise NotImplementedError
        if shared:
            raise NotImplementedError
        return ShellSession()